﻿//---------------------------------------------------------------------------
//
//   ExportSTL
//
//          Copyright(C) 2012-2021, tetraface Inc.
//
//     Sample of Export plug-in for exporting an STL (Stereolithography) 
//     file. Created DLL must be installed in "Plugins\Export" directory.
//
//    　STL(Stereolithography)形式を出力するエクスポートプラグインのサンプル。
//    　作成したDLLは"Plugins\Export"フォルダに入れる必要がある。
//
//---------------------------------------------------------------------------

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <Shlwapi.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include "MQPlugin.h"
#include "MQBasePlugin.h"
#include "MQ3DLib.h"
#include "MQSetting.h"
#include "MQWidget.h"
#include "Language.h"
#if __APPLE__
#include "../osx/StringUtil.h"
#include <CoreFoundation/CFBundle.h>
#include <Corefoundation/CFByteOrder.h>
#define MAX_PATH PATH_MAX
#define sprintf_s sprintf
#endif

#ifdef WIN32
#define FILEPATH_DELIMITER_CHAR L'\\'
#define FILEPATH_DELIMITER_L    L"\\"
#define FILEPATH_DELIMITER      "\\"
#else
#define FILEPATH_DELIMITER_CHAR L'/'
#define FILEPATH_DELIMITER_L    L"/"
#define FILEPATH_DELIMITER      "/"
#endif

inline static bool isPathSeparator(wchar_t character)
{
	return (character == L'\\' || character == L'/');
}

inline static wchar_t createPathSeparator(const std::wstring& path)
{
	for(const wchar_t *ptr = path.c_str() + path.length(); ptr > path.c_str(); ){
		--ptr;
		if(isPathSeparator(*ptr)){
			return *ptr;
		}
	}
	return FILEPATH_DELIMITER_CHAR; // return default separator
}

std::wstring extractDirectory(const std::wstring& src_path)
{
#ifdef WIN32
	wchar_t buf[MAX_PATH];
	wcscpy_s(buf, src_path.c_str());
	::PathRemoveFileSpecW(buf);
	return std::wstring(buf);
#else
	size_t len = src_path.length();
	for(const wchar_t *ptr = src_path.c_str() + len; ptr > src_path.c_str(); ){
		--ptr;
		if(isPathSeparator(*ptr)){
			std::wstring ret = src_path.substr(0, ptr - src_path.c_str());
			if(ret.length() > 0 && !isPathSeparator(ret[ret.length()-1])){
				ret += createPathSeparator(ret);
			}
			return ret;
		}
	}
	return std::wstring();
#endif
}

std::wstring combinePath(const std::wstring& base_dir, const std::wstring& cat_dir)
{
#ifdef _WIN32
	wchar_t buf[MAX_PATH];
	::PathCombineW(buf, base_dir.c_str(), cat_dir.c_str());
	return std::wstring(buf);
#endif
#if __APPLE__
	std::wstring ret;
	if(base_dir.length() > 0){
		ret += base_dir;
		if(base_dir[base_dir.length()-1] != L'\\' && base_dir[base_dir.length()-1] != L'/'){
			ret += L"/";
		}
	}
	ret += cat_dir;
	std::wstring::size_type p = ret.find_first_of(L'\\');
	while(p != std::wstring::npos){
		ret[p] = L'/';
		p = ret.find(L'\\', p+1);
	}
	return ret;
#endif
}

#ifdef _WIN32
static wchar_t s_DllPath[MAX_PATH];

//---------------------------------------------------------------------------
//  DllMain
//---------------------------------------------------------------------------
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	switch(ul_reason_for_call){
	case DLL_PROCESS_ATTACH:
		::GetModuleFileName((HMODULE)hModule, s_DllPath, _MAX_PATH);
		break;
	}

	// Nothing to do and return TRUE.
	//プラグインとしては特に必要な処理はないので、何もせずにTRUEを返す
    return TRUE;
}
#endif
#if __APPLE__
static std::wstring GetResourceDir()
{
	// Get a resource directory.
	CFBundleRef bundleRef = CFBundleGetBundleWithIdentifier(CFSTR("tetraface.ExportSTL"));
	if(bundleRef == NULL) return std::wstring();
	
	CFURLRef resourceUrl = CFBundleCopyResourcesDirectoryURL(bundleRef);
	char path[PATH_MAX+100];
	Boolean ret = CFURLGetFileSystemRepresentation(resourceUrl, TRUE, (UInt8*)path, PATH_MAX+100);
	CFRelease(resourceUrl);
	if(!ret) return std::wstring();
	
	return StringUtil::Utf8ToWide(path);
}
#endif


class ExportSTLPlugin : public MQExportPlugin
{
public:
	ExportSTLPlugin();
	~ExportSTLPlugin();

	void GetPlugInID(DWORD *Product, DWORD *ID) override;
	const char *GetPlugInName(void) override;
	const char *EnumFileType(int index) override;
	const char *EnumFileExt(int index) override;
	BOOL ExportFile(int index, const wchar_t *filename, MQDocument doc) override;

	MLanguage m_Language;

	void LoadResource();

private:
	BOOL SaveSTL(const wchar_t *filename, MQDocument doc);
};


MQBasePlugin *GetPluginClass()
{
	static ExportSTLPlugin plugin;
	return &plugin;
}

ExportSTLPlugin::ExportSTLPlugin()
{
}

ExportSTLPlugin::~ExportSTLPlugin()
{
}


//---------------------------------------------------------------------------
//  MQGetPlugInID
//    プラグインIDを返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
void ExportSTLPlugin::GetPlugInID(DWORD *Product, DWORD *ID)
{
	// Return a product(auther) ID and a plug-in ID.
	// They can be random values that do not overlap with other.
	// プロダクト名(制作者名)とIDを、全部で64bitの値として返す
	// 値は他と重複しないようなランダムなもので良い
	*Product = 0x56A31D20;
	*ID      = 0x3EA834CB;
}

//---------------------------------------------------------------------------
//  MQGetPlugInName
//    プラグイン名を返す。
//    この関数は[プラグインについて]表示時に呼び出される。
//---------------------------------------------------------------------------
const char *ExportSTLPlugin::GetPlugInName(void)
{
	// Return a plug-in name
	// プラグイン名
	return "Stereolithography Exporter   Copyright(C) 2012-2021, tetraface Inc.";
}

//---------------------------------------------------------------------------
//  MQEnumFileType
//    入力または出力可能なファイルタイプを返す。
//    ファイルタイプは別名保存時のダイアログに表示される。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
const char *ExportSTLPlugin::EnumFileType(int index)
{
	switch(index){
	case 0: return "Stereolithography (*.stl)";
	}
	return NULL;
}

//---------------------------------------------------------------------------
//  MQEnumFileExt
//    入力または出力可能な拡張子を返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
const char *ExportSTLPlugin::EnumFileExt(int index)
{
	switch(index){
	case 0: return "stl";
	}
	return NULL;
}

//---------------------------------------------------------------------------
//  MQExportFile
//    [別名で保存]で保存するときに呼び出される。
//---------------------------------------------------------------------------
BOOL ExportSTLPlugin::ExportFile(int index, const wchar_t *filename, MQDocument doc)
{
	switch(index){
	case 0: return SaveSTL(filename, doc);
	}
	return FALSE;
}


struct CreateDialogOptionParam
{
	ExportSTLPlugin *plugin;
	MQDialog *dialog;
	MQCheckBox *check_visible;
	
	bool visible_only;
};

static void CreateDialogOption(bool init, MQFileDialogCallbackParam *param, void *ptr)
{
	CreateDialogOptionParam *option = (CreateDialogOptionParam*)ptr;
	
	if(init)
	{
		option->dialog = new MQDialog(param->dialog_id);
		MQFrame parent(param->parent_frame_id);
	
		MQGroupBox *group = option->dialog->CreateGroupBox(&parent, option->plugin->m_Language.Search("Option"));
		option->check_visible = option->dialog->CreateCheckBox(group, option->plugin->m_Language.Search("VisibleOnly"));
		option->check_visible->SetChecked(option->visible_only);
	}
	else
	{
		option->visible_only = option->check_visible->GetChecked();
		delete option->dialog;
	}
}

void ExportSTLPlugin::LoadResource()
{
	if(m_Language.GetLanguage().length() == 0){
		std::wstring lang = GetSettingValue(MQSETTINGVALUE_LANGUAGE);
#ifdef WIN32
		std::wstring dir = extractDirectory(s_DllPath);
#else
		std::wstring dir = GetResourceDir();
#endif
		std::wstring path = combinePath(dir, L"ExportSTL.resource.xml");
		m_Language.Load(lang, path);
	}
}

// Export a file using a STL format
// STL形式で出力
BOOL ExportSTLPlugin::SaveSTL(const wchar_t *filename, MQDocument doc)
{
	LoadResource();

	CreateDialogOptionParam option;
	option.plugin = this;
	option.visible_only = false;

	MQSetting *setting = OpenSetting();
	if(setting != NULL){
		setting->Load("VisibleOnly", option.visible_only, option.visible_only);
		CloseSetting(setting);
	}

	// Show a dialog for converting axes
	// 座標軸変換用ダイアログの表示
	MQFileDialogInfo dlginfo;
	memset(&dlginfo, 0, sizeof(dlginfo));
	dlginfo.dwSize = sizeof(dlginfo);
	dlginfo.axis_x = MQFILE_TYPE_RIGHT;
	dlginfo.axis_y = MQFILE_TYPE_BACK;
	dlginfo.axis_z = MQFILE_TYPE_UP;
	dlginfo.softname = "";
	dlginfo.dialog_callback = CreateDialogOption;
	dlginfo.dialog_callback_ptr = &option;
	MQ_ShowFileDialog("STL Export", &dlginfo);

	setting = OpenSetting();
	if(setting != NULL){
		setting->Save("VisibleOnly", option.visible_only);
		CloseSetting(setting);
	}

	MQWaitCursorChanger wait_cursor(this);

	// Open a file.
	// ファイルを作成・オープンする
#if WIN32
	FILE *fh;
	errno_t ret_s = _wfopen_s(&fh, filename, L"wb");
	if(ret_s != 0){
		return FALSE;
	}
#else
	FILE *fh = fopen(MQEncoding::WideToUtf8(filename).c_str(), "wb");
	if(fh == NULL)
		return FALSE;
#endif

	// Create a temporal object for merging multiple objects.
	// 複数のオブジェクトを１つに合成するために、作業用オブジェクトを作成する
	MQObject obj = MQ_CreateObject();

	// Merge all objects to the temporal object.
	// 作業用オブジェクトにすべてのオブジェクトを合成
	int objcount = doc->GetObjectCount();
	for(int i=0; i<objcount; i++)
	{
		MQObject thisobj = doc->GetObject(i);
		if(thisobj == NULL) 
			continue;
		if(option.visible_only){
			if(thisobj->GetVisible() == 0)
				continue;
		}
		// Duplicate the object because the contents will be changed by a freezing and so on.
		// フリーズなどにより内容が変更されてしまうので、オブジェクトを複製しておく
		thisobj = thisobj->Clone();
		// Freeze attributes such as a patch.
		// 曲面などの属性をフリーズ
		thisobj->Freeze(MQOBJECT_FREEZE_ALL);
		// Merge to the temporal object.
		// 作業用オブジェクトに合成する
		obj->Merge(thisobj);
		// Delete the duplicated object.
		// 複製したオブジェクトはもう不要なので破棄
		thisobj->DeleteThis();
	}

	// Get a number of vertices and a number of faces.
	// 面と頂点の数を取得
	int face_count = obj->GetFaceCount();
	//int vert_count = obj->GetVertexCount();

	// Count a number of output triangles because a STL format cannot have 
	// quadrangles.
	// STL形式は三角形しか扱えないので、出力後の三角面の数を数える
	unsigned int triangle_count = 0;
	for(int face_index=0; face_index<face_count; face_index++)
	{
		int pt_num = obj->GetFacePointCount(face_index);
		if(pt_num >= 3){
			triangle_count += pt_num - 2;
		}
	}

	// Write a header description.
	// ファイルの先頭にデータを書き込み
	char header[80];
	memset(header, 0, sizeof(header));
	strcpy(header, "Created by Metasequoia.");
	fwrite(header, 80, 1, fh);

	// Write a number of triangles.
	// 三角形の数を出力
	fwrite(&triangle_count, 4, 1, fh);

	// Write triangles.
	// 三角形を出力
	for(int face_index=0; face_index<face_count; face_index++)
	{
		int pt_num = obj->GetFacePointCount(face_index);
		if(pt_num < 3) continue;

		// Get indices of vertices.
		// 頂点インデックスの配列を得る
		std::vector<int> index(pt_num);
		obj->GetFacePointArray(face_index, &(*index.begin()));

		int iter = pt_num - 2;
		std::vector<int> indices(iter*3);
		if(pt_num == 3){
			indices[0] = 0;
			indices[1] = 1;
			indices[2] = 2;
		}else{
			std::vector<MQPoint> pts(pt_num);
			for(int i=0; i<pt_num; i++){
				pts[i] = obj->GetVertex(index[i]);
			}
			doc->Triangulate(&(*pts.begin()), pt_num, &(*indices.begin()), iter*3);
		}

		for(int i=0; i<iter; i++)
		{
			float val[3];
			MQPoint pts[3];

			// Get positions of vertices.
			// 頂点の位置を取得
			pts[0] = obj->GetVertex(index[indices[i*3+0]]);
			pts[1] = obj->GetVertex(index[indices[i*3+1]]);
			pts[2] = obj->GetVertex(index[indices[i*3+2]]);

			// Convert an axis coordinate of the vertex.
			// 座標軸をダイアログで指定したものに変換
			MQ_ExportAxis(&dlginfo, pts, 3);

			// Calculate a normal vector.
			// 法線ベクトルを計算
			MQPoint nrm;
			if(dlginfo.hand_type == MQFileDialogInfo::HAND_RIGHT){
				nrm = GetNormal(pts[2], pts[1], pts[0]);
			}else{
				nrm = GetNormal(pts[0], pts[1], pts[2]);
			}

			// Write the normal vector.
			// 法線ベクトルを出力
			val[0] = nrm.x;
			val[1] = nrm.y;
			val[2] = nrm.z;
			fwrite(val, 4, 3, fh);

			// Write the vertices.
			// 頂点座標を出力
			int i0,i1,i2;
			if(dlginfo.hand_type == MQFileDialogInfo::HAND_RIGHT){
				i0 = 0;
				i1 = 1;
				i2 = 2;
			}else{
				i0 = 2;
				i1 = 1;
				i2 = 0;
			}
			val[0] = pts[i0].x;
			val[1] = pts[i0].y;
			val[2] = pts[i0].z;
			fwrite(val, 4, 3, fh);
			val[0] = pts[i1].x;
			val[1] = pts[i1].y;
			val[2] = pts[i1].z;
			fwrite(val, 4, 3, fh);
			val[0] = pts[i2].x;
			val[1] = pts[i2].y;
			val[2] = pts[i2].z;
			fwrite(val, 4, 3, fh);

			// Write an unused attribute
			static const unsigned char dummy[2] = {0,0};
			fwrite(dummy, 2, 1, fh);
		}
	}

	// Delete the temporal object.
	// 作業用オブジェクトを破棄
	obj->DeleteThis();

	// Close a file.
	// ファイルをクローズ
	int ret = fclose(fh);

	return (ret == 0);
}

