#include <Foundation/Foundation.h>
#include "StringUtil.h"
#include <codecvt>
#include <locale>

namespace StringUtil {

	static NSStringEncoding CodePageToNSStringEncoding(int codepage)
	{
		switch(codepage){
			case kCodePage_ShiftJis:
				return NSShiftJISStringEncoding;
			case kCodePage_WindowsCentralEuropean:
				return NSWindowsCP1250StringEncoding;
			case kCodePage_WindowsCyrillic:
				return NSWindowsCP1251StringEncoding;
			case kCodePage_WindowsLatin1:
				return NSWindowsCP1252StringEncoding;
			case kCodePage_WindowsGreek:
				return NSWindowsCP1253StringEncoding;
			case kCodePage_WindowsTurkish:
				return NSWindowsCP1254StringEncoding;
			default:
			{
				CFStringEncoding cfenc = CFStringConvertWindowsCodepageToEncoding(codepage);
				return CFStringConvertEncodingToNSStringEncoding(cfenc);
			}
			case kCodePage_Default:
			{
				CFStringEncoding cfenc = CFStringGetSystemEncoding();
				return CFStringConvertEncodingToNSStringEncoding(cfenc);
			}
		}
	}
	
	int GetDefaultCodePage()
	{
		CFStringEncoding cfenc = CFStringGetSystemEncoding();
		int codepage = CFStringConvertEncodingToWindowsCodepage(cfenc);
		return codepage;
	}
	
	std::string WideToCodePageString(const std::wstring& mstr, int codepage)
	{
		NSString *nsstr = [[NSString alloc] initWithBytes:mstr.c_str() length:mstr.length()*sizeof(wchar_t) encoding:NSUTF32LittleEndianStringEncoding];
		NSStringEncoding enc = CodePageToNSStringEncoding(codepage);
		NSData* data = [nsstr dataUsingEncoding:enc allowLossyConversion:YES];
		const char *chars = static_cast<const char*>([data bytes]);
		return std::string(chars, (size_t)[data length]);
	}

	std::wstring CodePageStringToWide(const std::string& ansistr, int codepage)
	{
		if(ansistr.length() == 0)
			return std::wstring();
		NSData *sjisData = [NSData dataWithBytes:ansistr.c_str() length:ansistr.length()+1];
		NSStringEncoding enc = CodePageToNSStringEncoding(codepage);
		NSString *nsstr = [[NSString alloc] initWithData:sjisData encoding:enc];
		const wchar_t *chars = (const wchar_t*)[nsstr cStringUsingEncoding:NSUTF32StringEncoding];
		return std::wstring(chars);
	}

	std::string WideToShiftJis(const std::wstring& wstr)
	{
		NSString *nsstr = [[NSString alloc] initWithBytes:wstr.c_str() length:wstr.length()*sizeof(wchar_t) encoding:NSUTF32LittleEndianStringEncoding];
		NSData* sjisData = [nsstr dataUsingEncoding:NSShiftJISStringEncoding allowLossyConversion:YES];
		const char *chars = static_cast<const char*>([sjisData bytes]);
		return std::string(chars, (size_t)[sjisData length]);
	}

	std::wstring ShiftJisToWide(const std::string& ansistr)
	{
		NSData *sjisData = [NSData dataWithBytes:ansistr.c_str() length:ansistr.length()+1];
		NSString *nsstr = [[NSString alloc] initWithData:sjisData encoding:NSShiftJISStringEncoding];
		const char *chars = [nsstr cStringUsingEncoding:NSUTF8StringEncoding];
		return Utf8ToWide(chars);
	}

	std::wstring Utf8ToWide(const std::string& utf8str)
	{
		try {
			std::wstring_convert<std::codecvt_utf8<wchar_t>, wchar_t> conv;
			return conv.from_bytes(utf8str);
		}catch(...){
		}
		return std::wstring();
	}

	std::string WideToUtf8(const std::wstring& wstr)
	{
		try {
			std::wstring_convert<std::codecvt_utf8_utf16<wchar_t>, wchar_t> conv;
			return conv.to_bytes(wstr);
		}catch(...){
		}
		return std::string();
	}
	
}

